require "util"

-- set this demo_print flag to true to cause 
-- procedures to 
-- print results *every* time through.  
-- They will also wait for some input; press <return>.  
-- This causes a *lot* of output on big runs.  

demo_print = false 

--[[	

*** BUBBLE SORT command ***

When applied to a given array { 4, 3, 5, ...}, it repeatedly bubbles
out-of-order elements toward their correct positions, ending when
there are no more out-of-order pairs.

Note:  This procedure will permanently change the order of entries in
the array.  Copy the array before giving it if you want to keep the
original!

--]]


function bubble_sort (a) 
   local a_length = #a		-- Lua notation for array length is # 
   local still_active = true 	-- flag to decide when to stop 

   while still_active do 
      still_active = false 	-- will any work be done this time? 

      for i = 1, a_length-1 do 
	 if a[i+1]<a[i] 	-- Is this pair out of order? 
	 then 
	    -- Yes, you can swap values like this in Lua: 
	    a[i], a[i+1] = 
	       a[i+1], a[i]	

	    still_active = true	-- Did work, so dont stop yet! 
	 end
      end 
      if demo_print then util.print_array(a); io.read() end 
   end
   return a
end 


--[[	

*** insertion sort command ***
   insertion_sort 

for each slot in the array given, we place the value there in the
right position relative to all the values in earlier slots.  

Note:  This procedure will permanently change the order of entries in
the array.  Copy the array before giving it if you want to keep the
original!

--]]

function insertion_sort(a)

   for i = 2, #a do 		-- the target for each step will be a[i]
      local entry = a[i] 

	 -- the sorted part at any time 
	 -- is a[1] .. a[i-1].

      for j = i-1, 0, -1 do 	-- count down from position i-1

	 -- we will insert the target in the right spot, 
	 -- and slide the part above that up.

	 if j == 0 or entry >= a[j] 
	 then 			-- put the entry above a[j] 
	    a[j+1] = entry 
	    break 		-- bail out of rest of loop "for j"
	 else 			
	    -- slide this a[j] up to the next position 
	    a[j+1] = a[j]	
	 end 
      end 
      if demo_print then util.print_array(a); io.read() end 
   end 
   return a
end 

--[[
***   merge sort  ***

heres an implementation of merge sort, 
organized as three functions.

the first is a utility function to do copying.  

the second merges two already sorted arrays.    

the third does the wishful thinking, namely 
it applies itself recursively to the two halves 
of its argument, and merges the results for the 
two halves to build the full array.  

--]] 


-- copy from src to target 
-- from lower to upper *inclusive*. 

-- return the max entry found, and the index 
-- for a slot to place an upper bound
-- to serve as a sentinel.

function copy_range_with_slot(src, lower, upper, target) 
   local max = src[lower] 

   for i = 0, upper-lower do	-- loop will run upper-lower+1 times
      local entry = src[i+lower] -- next entry 
      target[i+1] = entry	 -- copy it
      max = math.max(max, entry) -- update max 
   end 
   return max, upper-lower+2	-- the latter is the first unused slot 
end 

--[[ 

merge: 

assume that a is an array, 
and the entries from lower to mid are
already in order,
and the entries from mid+1 to upper are
already in order.

modify the entries in a so that
they are all ordered between lower and upper.  

that is, make sure the entries in 
the lower part go in the right spots
versus the entries in the upper part.

--]]



function merge_into_a(a, lower, mid, upper, spare1, spare2) 
   local max = a[lower] 

   -- first copy contents to spare arrays

   local max1, slot1 =
      copy_range_with_slot(a, lower, mid, spare1) 
   local max2, slot2 = 
      copy_range_with_slot(a, mid+1, upper, spare2) 

   -- fill in both sentinels, greater than any real entry

   spare1[slot1] = math.max(max1,max2)+1 
   spare2[slot2] = math.max(max1,max2)+1

   local j, k= 1, 1

   for i = lower, upper do 
      if spare1[j]<spare2[k] 	-- use smaller value
      then 
	 a[i] = spare1[j] 
	 j = j+1 		-- increment index in array used 
      else 
	 a[i] = spare2[k] 
	 k = k+1 
      end 
   end 
   if demo_print then util.print_array(a); print(lower, upper); io.read() end
end 

-- merge_sort(a) 
-- splits the array a unless it is of length 0,1
-- and recursively merge_sorts each half
-- then it uses merge to combine the sorted halves.

	 
function merge_sort(a, lower, upper, spare1, spare2) 
   if upper<=lower then return end 
   
   local mid = math.floor((lower+upper) / 2)

   merge_sort(a, lower, mid,	-- sort lower half recursively 
      spare1, spare2)
   merge_sort(a, mid+1, upper, 	-- sort upper half recursively 
      spare1, spare2)
   merge_into_a(a, lower, mid, upper, -- merge the two sorted halves 
		spare1, spare2)
end 
      
--[[

Top level entry to merge sort.  

Note:  This procedure will permanently change the order of entries in
the array.  Copy the array before giving it if you want to keep the
original!

--]]


function ms_top(a) 
   merge_sort(a,1,#a,		-- sort the whole array 
	      {},{})		-- with new empty tables as spare space 
end 



